from django.conf import settings
from pymongo import MongoClient


def mongo_db():

    return MongoClient(
        getattr(settings, "MONGODB_URL", "mongodb://localhost:27017")
    )[getattr(settings, "MONGODB_DATABASE", "archive")]




from copy import deepcopy
import re

from django.db.models.sql.constants import QUERY_TERMS
from bson.objectid import ObjectId

OPERATORS_MAP = {
    'exact': lambda val: val,
    'gt': lambda val: {'$gt': val},
    'gte': lambda val: {'$gte': val},
    'lt': lambda val: {'$lt': val},
    'lte': lambda val: {'$lte': val},
    'in': lambda val: {'$in': val},
    'range': lambda val: {'$gte': val[0], '$lte': val[1]},
    'isnull': lambda val: None if val else {'$ne': None},
    'iexact': lambda val: re.compile('^%s$' % re.escape(val), re.IGNORECASE),
    'startswith': lambda val: re.compile('^%s' % re.escape(val)),
    'istartswith': lambda val: re.compile('^%s' % re.escape(val), re.IGNORECASE),
    'endswith': lambda val: re.compile('%s$' % re.escape(val)),
    'iendswith': lambda val: re.compile('%s$' % re.escape(val), re.IGNORECASE),
    'contains': lambda val: re.compile('%s' % re.escape(val)),
    'icontains': lambda val: re.compile('%s' % re.escape(val), re.IGNORECASE),
    'regex': lambda val: re.compile(val),
    'iregex': lambda val: re.compile(val, re.IGNORECASE)
}


class MongoFilter(object):
    select_related = False
    query_terms = QUERY_TERMS

    def __init__(self, filter_terms, order_by=None, model=None):
        self.filter_terms = filter_terms
        if order_by is None:
            order_by = []
        self.order_by = order_by
        self.where = None
        self.model = model

    def clone(self):
        return MongoFilter(deepcopy(self.filter_terms), order_by=deepcopy(self.order_by),
                           model=self.model)

    def add_filters(self, extra_filters):
        for name, value in extra_filters.items():
            # If you re-filter on the same column this will replace what you filtered before
            self.filter_terms[name] = value

    @property
    def is_empty(self):
        return len(self.filter_terms) == 0

    def as_mongo_filter(self):
        # Convert this to a mongo filter
        mongo_filter = {}
        for name, value in self.filter_terms.items():
            fields = name.split("__")
            op = None
            if len(fields) > 1 and fields[-1] in OPERATORS_MAP.keys():
                op = fields[-1]

                if fields[0] == ["id", "pk"] and op in ["exact", "iexact"]:
                    value = ObjectId(value)
                elif fields[0] == ["id", "pk"] and op in ["in"]:
                    value = map(lambda x: ObjectId(x), value)
                fields = fields[:-1]

            if fields[0] in ["id", "pk"]:
                fields[0] = "_id"
                if op is None:
                    value = ObjectId(value)
            elif len(fields) == 1:
                if isinstance(value, (list, tuple)):
                    value = map(
                         lambda x: self.model._meta.get_field(fields[0]).get_prep_value(x),
                         value
                     )
                else:
                    value = self.model._meta.get_field(fields[0]).get_prep_value(value)

            if op:
                value = OPERATORS_MAP[op](value)

            mongo_filter[".".join(fields)] = value

        return mongo_filter



import sys

from django.db.models.base import ModelBase
from django.db.models.options import Options
from django.db.models import FieldDoesNotExist, Field

from .manager import MongoManager
from .utils import mongo_db


class MongoModelMeta(ModelBase):

    def __new__(cls, name, bases, attrs):
        new_class = super(ModelBase, cls).__new__(cls, name, bases, attrs)
        new_class.objects.model = new_class
        new_class._default_manager.model = new_class

        meta = getattr(new_class, 'Meta')
        if getattr(meta, 'app_label', None) is None:
            # Figure out the app_label by looking one level up.
            # For 'django.contrib.sites.models', this would be 'sites'.
            model_module = sys.modules[new_class.__module__]
            kwargs = {"app_label": model_module.__name__.split('.')[-2]}
        else:
            kwargs = {"app_label": meta.app_label}

        new_class._meta = Options(meta, **kwargs)
        new_class._meta.object_name = new_class.__name__
        new_class._meta.module_name = new_class.__name__.lower()
        new_class._meta.concrete_model = new_class

        if not hasattr(new_class.Meta, "verbose_name"):
            new_class._meta.verbose_name = new_class.__name__.lower()
        if not hasattr(new_class.Meta, "verbose_name_plural"):
            new_class._meta.verbose_name_plural = new_class._meta.verbose_name + "s"

        class Pk(Field):
            name = "pk"
            attname = "id"

        new_class._meta.pk = Pk()
        new_class._meta.id = Pk()
        new_class._meta.id.name = "id"
        #new_class._meta.fields.append(new_class._meta.pk)
        #new_class._meta.fields.append(new_class._meta.id)

        for attr, attr_value in attrs.items():
            if hasattr(attr_value, 'contribute_to_class'):
                attr_value.name = attr
                setattr(new_class, attr, None)
                attr_value.name = attr
                new_class._meta.fields.append(attr_value)

        return new_class


class MongoModel(object):
    __metaclass__ = MongoModelMeta

    _default_manager = MongoManager()

    class DoesNotExist(Exception):
        pass

    class Meta(object):
        pass

    fields = []
    objects = MongoManager()
    _meta = None

    def __init__(self, **kwargs):
        self._id = None
        self.__dict__.update(kwargs)

    @property
    def id(self):
        return str(self._id)

    @property
    def pk(self):
        return str(self._id)

    def delete(self):
        if self._id is None:
            return
        self.objects.filter(_id=self._id).delete()

    @property
    def _message(self):
        message = {}
        for field in self._meta.fields:
            message[field.name] = getattr(self, field.name, None)
        return message

    @classmethod
    def collection_name(cls):
        return getattr(cls.Meta, "db_table", cls.__class__.__name__.lower())

    @classmethod
    def collection(cls):
        return mongo_db()[cls.collection_name()]

    def save(self):
        # Set defaults and make everything the right object
        for field in self._meta.fields:
            value = getattr(self, field.name, None)
            if value is None:
                setattr(self, field.name, field.get_default())
            else:
                setattr(self, field.name, field.get_prep_value(value))

        if self._id is None:
            self._id = self.collection().insert(self._message)
        else:
            self.objects.filter(_id=self._id).update(**self._message)

    def serializable_value(self, field_name):
        """
        Returns the value of the field name for this instance. If the field is
        a foreign key, returns the id value, instead of the object. If there's
        no Field object with this name on the model, the model attribute's
        value is returned directly.

        Used to serialize a field's value (in the serializer, or form output,
        for example). Normally, you would just access the attribute directly
        and not use this method.
        """
        try:
            field = self._meta.get_field_by_name(field_name)[0]
        except FieldDoesNotExist:
            return getattr(self, field_name)
        return getattr(self, field.attname)


from pymongo import ASCENDING, DESCENDING

from .filter import MongoFilter


class MongoQuerySet(object):

    def __init__(self, model=None, cursor=None):
        self.model = model
        self.cursor = cursor
        self.query = MongoFilter({}, model=model)

    def get_cursor(self):
        if self.cursor is None:
            raise Exception("No cursor - Query set wasn't run against anything")
        return self.cursor

    def __iter__(self):
        return self

    def next(self):
		# Needed when mocking mongo
        if isinstance(self.cursor, list):
            self.cursor = iter(self.cursor)
        return self.doc_to_instance(self.cursor.next())

    def filter(self, *args, **kwargs):
        # Convert filter to mongo format
        self.query.add_filters(kwargs)
        self.cursor = self.model.collection().find(self.query.as_mongo_filter())
        return self

    def get(self, **kwargs):
        try:
            return self.filter(**kwargs)[0]
        except IndexError:
            raise self.model.DoesNotExist()

    def delete(self, delete_all=False):
        if self.query.is_empty and not delete_all:
            raise Exception("If you want to delete everything in collection() pass delete_all=True")
        self.model.collection().find(self.query.as_mongo_filter())

    def count(self):
        return self.get_cursor().count()

    def clone(self):
        return self._clone()

    def _clone(self):
        """This method seems pointless but is called by libraries like tastypie"""
        cloned_cursor = None
        if self.cursor:
            cloned_cursor = self.cursor.clone()
        cloned_queryset = MongoQuerySet(model=self.model, cursor=cloned_cursor)
        cloned_queryset.query = self.query.clone()
        return cloned_queryset

    def doc_to_instance(self, doc):
        instance = self.model()
        for name, value in doc.items():
            setattr(instance, name, value)
        return instance

    def update(self, **kwargs):
        # Todo: handle timezone aware datetime fields
        updated_fields = {}
        for name, value in kwargs.items():
            updated_fields[".".join(name.split("__"))] = value
        self.model.collection().update(
            self.query.as_mongo_filter(),
            {"$set": updated_fields}
        )

    def __getitem__(self, index):
        clone = self.clone()

        if isinstance(index, slice):
            clone.cursor = clone.cursor.__getitem__(index)
            return clone

        if isinstance(index, (int, long)):
            return self.doc_to_instance(clone.cursor.__getitem__(index))

        raise TypeError("index %r cannot be applied to Cursor instances" % index)

    def order_by(self, *args):
        self.query.order_by = args
        sort_list = []
        for arg in args:
            direction = ASCENDING
            if arg.startswith("-"):
                direction = DESCENDING
                arg = arg[1:]  # Strip the -
            sort_list.append((arg, direction))
        if len(sort_list) == 0:
            return self
        self.cursor = self.get_cursor().sort(sort_list)
        return self

    def all(self):
        self.cursor = self.model.collection().find()
        return self

    def __len__(self):
        return self.count()



from .queryset import MongoQuerySet


class MongoManager(object):

    def __init__(self, model=None):
        self.model = model

    def filter(self, *args, **kwargs):
        return MongoQuerySet(model=self.model).filter(*args, **kwargs)

    def all(self):
        return MongoQuerySet(model=self.model).all()

    def get(self, **kwargs):
        return MongoQuerySet(model=self.model).get(**kwargs)

    def create(self, **kwargs):
        instance = self.model(**kwargs)
        instance.save()
        return instance

    def count(self):
        return self.model.collection().count()

    def get_query_set(self):
        return MongoQuerySet(model=self.model)



import os
from setuptools import setup
import ConfigParser


def load_metadata(ini_file):
    config = ConfigParser.ConfigParser()
    read_files = config.read(ini_file)
    if len(read_files) == 0:
        raise Exception("Failed to read %s" % ini_file)
    meta = {}
    for item in config.options("meta"):
        meta[item] = config.get("meta", item)
    return meta

METADATA_PATH = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'gentle_django_mongo/METADATA')
META = load_metadata(METADATA_PATH)


def read(fname):
    try:
        return open(os.path.join(os.path.dirname(__file__), fname)).read()
    except:
        return ""

setup(
    name=META["name"],
    version=META["version"],
    author="Sebastian Phelps",
    author_email="sebastian.phelps@gmail.com",
    description=META["description"],
    keywords="gentle_django_mongo mongo django model",
    packages=['gentle_django_mongo'],
    include_package_data=True,
    package_data={'gentle_django_mongo': ['METADATA', ]},
    long_description=read('README.md'),
    requires=[
        'django (>=1.4.1)',
        'pymongo (>=2.7.1)'
    ],
    install_requires=[
        'django>=1.4.1',
        'pymongo>=2.7.1'
    ]
)
